import pathlib
from typing import Any, List, Optional, Tuple
from unittest import mock

from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem

from gcs_migration.commands.create import create
from gcs_migration.models import MigrationPlan
from gcs_migration.parsers import (
    BlackPearlMapEntry,
    GCSConfig,
    GridFTPOptionParser,
    S3Config,
    S3MapfileParser,
)
from gcs_migration.util import AuthorizeSharing


def test_create(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
    blackpearl_endpoint: str,
    blackpearl_access_id_filename: str,
    blackpearl_map_entries: List[BlackPearlMapEntry],
    mock_s3_config: S3Config,
    mock_s3_mapfile: S3MapfileParser,
    mock_logged_in: None,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.BlackPearlConfig"
    ) as mock_blackpearl_config, mock.patch(
        "gcs_migration.models.storage_gateway.BlackPearlAccessIdFileParser"
    ) as mock_bp_access_id_parser, mock.patch(
        "gcs_migration.commands.create.S3Config", return_value=mock_s3_config
    ), mock.patch(
        "gcs_migration.models.storage_gateway.S3MapfileParser",
        return_value=mock_s3_mapfile,
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan
        mock_blackpearl_config.return_value.endpoint = blackpearl_endpoint
        mock_blackpearl_config.return_value.access_id_file = (
            blackpearl_access_id_filename
        )
        mock_access_id_parser = mock_bp_access_id_parser.return_value
        mock_access_id_parser.map_entries = {
            k.username: k for k in blackpearl_map_entries
        }
        mock_bp_access_id_parser.return_value.__contains__.side_effect = (
            lambda x: x in mock_access_id_parser.map_entries
        )
        mock_bp_access_id_parser.return_value.__getitem__.side_effect = (
            lambda x: mock_access_id_parser.map_entries[x]
        )

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 0
        loaded_plan = MigrationPlan.load(tmp_path)

        assert (
            loaded_plan.gsi_authz_configuration
            == migration_plan_cmd_create.gsi_authz_configuration
        )
        assert loaded_plan.storage_gateway == migration_plan_cmd_create.storage_gateway
        assert (
            loaded_plan.mapped_collection == migration_plan_cmd_create.mapped_collection
        )
        assert (
            loaded_plan.guest_collections == migration_plan_cmd_create.guest_collections
        )

        assert loaded_plan == migration_plan_cmd_create
        assert loaded_plan == ret_migration_plan


def test_create_v5_endpoint(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    v5_endpoint_uuid: str,
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        gcs_config.endpoint_id = v5_endpoint_uuid
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "is a GCS v5 endpoint" in result.output


def test_create_v5_node(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        gridftp_config_parser.options["version_tag"] = "GCS-5.4.50"
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "GridFTP server configuration is for a GCSv5 endpoint" in result.output


def test_create_v5_collection(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    v5_collection_uuid: str,
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        gcs_config.endpoint_id = v5_collection_uuid
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "is a GCS v5 collection" in result.output


def test_requires_logged_in(mock_not_logged_in: None) -> None:
    runner = CliRunner()
    result = runner.invoke(create, catch_exceptions=False)
    assert "you will have to login" in result.output
