import os
import pathlib
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple
from typing_extensions import Protocol
from unittest import mock

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem

from gcs_migration.parsers import GridFTPOptionParser
from gcs_migration.parsers.gridftp_options import options as all_gridftp_options

from .gridmap import GridmapFileFixtureType


class GridFTPConfigPolicy(Protocol):
    options: Dict[str, Any]
    filename: str

    def __str__(self) -> str:
        ...


@pytest.fixture
def gridftp_config_sharing_policies(sharing_state_dir: str) -> GridFTPConfigPolicy:
    class SharingPolicies:
        options = {
            "sharing_dn": "/C=US/O=Globus Consortium/OU=Globus Online/OU=Transfer User/CN=__transfer__",
            "sharing_users_allow": "allow1,allow2",
            "sharing_users_deny": "deny1,deny2,deny3",
            "sharing_groups_allow": "gallow1,gallow2,gallow3",
            "sharing_groups_deny": "gdeny1",
            "sharing_rp": "R/home/share,RW/project/share,N/noshare",
            "sharing_state_dir": sharing_state_dir,
        }
        filename = "sharing_policies"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return SharingPolicies()


@pytest.fixture
def gridftp_config_restrict_paths() -> "GridFTPConfigPolicy":
    class PathRestrictions:
        options = {"restrict_paths": "R/home,RW/projects,N/"}
        filename = "path_restrictions"

        def __str__(self) -> str:
            return """\
restrict_paths R/home,RW/projects,N/
"""

    return PathRestrictions()


@pytest.fixture
def gridftp_conf(fs: FakeFilesystem) -> GridFTPConfigPolicy:
    name = "gridftp.conf"

    class GridftpConf:
        options = {"port": 2811}
        filename = str(name)

        def __str__(self) -> str:
            return f"""\
# globus-gridftp-server configuration file

# this is a comment

# option names beginning with '$' will be set as environment variables, e.g.
# $GLOBUS_ERROR_VERBOSE 1
# $GLOBUS_TCP_PORT_RANGE 50000,51000

# port
port {self.options["port"]}
"""

    conf = GridftpConf()

    with open(name, "w") as f:
        print(str(conf), file=f)

    return conf


@pytest.fixture
def gridftp_config_gcs() -> GridFTPConfigPolicy:
    class GlobusConnectServer:
        options = {
            "version_tag": "GCS-4.0.62",
            "usage_stats_id": "GCS-4.0.62+centos-7-x",
            "port_range": "50000,51000",
            "data_interface": "192.168.1.1",
        }

        filename = "globus-connect-server"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return GlobusConnectServer()


@pytest.fixture
def gridftp_config_authorization_myproxy(
    myproxy_gsi_authz_conf: str, gridmap_file: GridmapFileFixtureType
) -> GridFTPConfigPolicy:
    class Authorization:
        options = {
            "$GSI_AUTHZ_CONF": myproxy_gsi_authz_conf,
            "$GRIDMAP": gridmap_file[0],
        }

        filename = "globus-connect-server-authorization"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return Authorization()


@pytest.fixture
def gridftp_config_credential() -> GridFTPConfigPolicy:
    class Credential:
        options = {
            "$X509_USER_CERT": "/var/lib/globus-connect-server/grid-security/hostcert.pem",
            "$X509_USER_KEY": "/var/lib/globus-connect-server/grid-security/hostkey.pem",
        }
        filename = "globus-connect-server-credential"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return Credential()


@pytest.fixture
def gridftp_config_logging() -> GridFTPConfigPolicy:
    class Logging:
        options = {
            "log_single": "/var/log/gridftp.log",
            "log_level": "ERROR,WARN",
        }
        filename = "globus-connect-server-logging"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return Logging()


@pytest.fixture(params=[None, "globuss3"])
def s3_process_user(request: Any) -> Optional[str]:
    return request.param


@pytest.fixture(params=[None, "globusblackpearl"])
def blackpearl_process_user(request: Any) -> Optional[str]:
    return request.param


@pytest.fixture
def gridftp_config_s3_dsi() -> GridFTPConfigPolicy:
    class S3DSI:
        options = {
            "load_dsi_module": "s3",
            "threads": 2,
        }

        filename = "globus-connect-server-s3"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return S3DSI()


@pytest.fixture
def gridftp_config_s3_dsi_process_user(
    s3_process_user: Optional[str],
) -> GridFTPConfigPolicy:
    class S3DSI:
        options = {
            "load_dsi_module": "s3",
            "threads": 2,
        }
        if s3_process_user:
            options["process_user"] = s3_process_user

        filename = "globus-connect-server-s3"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    return S3DSI()


@pytest.fixture
def gridftp_config_blackpearl_dsi_default_path(
    request: Any,
    blackpearl_default_config_file_path: str,
) -> Generator[GridFTPConfigPolicy, None, None]:
    class BlackPearlDSI:
        def __init__(self):
            self.options = {
                "load_dsi_module": "blackpearl",
                "threads": 2,
            }

            self.filename = "globus-connect-server-blackpearl"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    yield BlackPearlDSI()


@pytest.fixture
def gridftp_config_blackpearl_dsi_custom_path(
    request: Any,
    blackpearl_custom_config_file_path: str,
) -> Generator[GridFTPConfigPolicy, None, None]:
    class BlackPearlDSI:
        def __init__(self):
            self.options = {
                "load_dsi_module": "blackpearl",
                "threads": 2,
                "$BLACKPEARL_DSI_CONFIG_FILE": blackpearl_custom_config_file_path,
            }
            self.filename = "globus-connect-server-blackpearl"

        def __str__(self) -> str:
            return "".join(f"{k} {v}\n" for k, v in self.options.items())

    yield BlackPearlDSI()


GridFTPConfigFixture = Tuple[os.PathLike, Dict[str, Any]]
GridFTPConfigFixtureGenerator = Callable[
    [List[GridFTPConfigPolicy]], GridFTPConfigFixture
]


@pytest.fixture
def gridftp_config_d(fs: FakeFilesystem) -> GridFTPConfigFixtureGenerator:
    def _gridftp_config_d(configs: List[GridFTPConfigPolicy]) -> GridFTPConfigFixture:
        gridftp_d = pathlib.Path("gridftp.d")
        if not os.path.exists(gridftp_d):
            os.makedirs(gridftp_d, 0o0755)

        options: Dict[str, Any] = {}
        for config in configs:
            with open(gridftp_d / config.filename, "w") as f:
                f.write(str(config))
            options = {**options, **config.options}

        return (gridftp_d, options)

    return _gridftp_config_d


@pytest.fixture
def gridftp_config(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_authorization_myproxy: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
) -> GridFTPConfigFixture:
    configs = [
        gridftp_config_gcs,
        gridftp_config_authorization_myproxy,
        gridftp_config_credential,
        gridftp_config_logging,
    ]
    config_dir, config_dir_options = gridftp_config_d(configs)
    config_data = {**gridftp_conf.options, **config_dir_options}

    return config_dir, config_data


@pytest.fixture
def gridftp_config_s3(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_authorization_myproxy: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
    gridftp_config_s3_dsi: GridFTPConfigPolicy,
) -> GridFTPConfigFixture:
    configs = [
        gridftp_config_gcs,
        gridftp_config_authorization_myproxy,
        gridftp_config_credential,
        gridftp_config_logging,
        gridftp_config_s3_dsi,
    ]
    config_dir, config_dir_options = gridftp_config_d(configs)
    config_data = {**gridftp_conf.options, **config_dir_options}

    return config_dir, config_data


@pytest.fixture
def gridftp_config_blackpearl(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_authorization_myproxy: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
    gridftp_config_blackpearl_dsi_default_path: GridFTPConfigPolicy,
) -> GridFTPConfigFixture:
    configs = [
        gridftp_config_gcs,
        gridftp_config_authorization_myproxy,
        gridftp_config_credential,
        gridftp_config_logging,
        gridftp_config_blackpearl_dsi_default_path,
    ]
    config_dir, config_dir_options = gridftp_config_d(configs)
    config_data = {**gridftp_conf.options, **config_dir_options}

    return config_dir, config_data


@pytest.fixture
def fake_gridftp_config_parser_creator() -> Callable[
    [GridFTPConfigFixture], GridFTPOptionParser
]:
    def _fake_gridftp_config_parser_creator(
        fixtures: GridFTPConfigFixture,
    ) -> GridFTPOptionParser:
        with mock.patch(
            "gcs_migration.parsers.gridftp_option_parser.ctypes"
        ), mock.patch(
            "gcs_migration.parsers.gridftp_option_parser.Process"
        ), mock.patch(
            "gcs_migration.parsers.gridftp_option_parser.Queue"
        ) as mock_queue:
            options = fixtures[1]
            for option in all_gridftp_options.values():
                if option.name not in options:
                    options[option.name] = option.default
                elif isinstance(options[option.name], os.PathLike):
                    options[option.name] = str(options[option.name])
            mock_queue.return_value.get = options
            mock_parser = GridFTPOptionParser()
            mock_parser.options = options

        return mock_parser

    return _fake_gridftp_config_parser_creator


@pytest.fixture
def gridftp_config_parser(
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], Generator[GridFTPOptionParser, None, None]
    ],
    connector: str,
    gridftp_config: GridFTPConfigFixture,
    gridftp_config_s3: GridFTPConfigFixture,
    gridftp_config_blackpearl: GridFTPConfigFixture,
) -> Generator[GridFTPOptionParser, None, None]:
    if connector == "file":
        return fake_gridftp_config_parser_creator(gridftp_config)
    elif connector == "s3":
        return fake_gridftp_config_parser_creator(gridftp_config_s3)
    elif connector == "blackpearl":
        return fake_gridftp_config_parser_creator(gridftp_config_blackpearl)
    else:
        raise AssertionError(f"Unsupported connector {connector} in fixture")


@pytest.fixture
def gridftp_config_parser_s3(
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], Generator[GridFTPOptionParser, None, None]
    ],
    gridftp_config_s3: GridFTPConfigFixture,
) -> Generator[GridFTPOptionParser, None, None]:
    return fake_gridftp_config_parser_creator(gridftp_config_s3)


@pytest.fixture
def gridftp_config_parser_blackpearl(
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], Generator[GridFTPOptionParser, None, None]
    ],
    gridftp_config_blackpearl: GridFTPConfigFixture,
) -> Generator[GridFTPOptionParser, None, None]:
    return fake_gridftp_config_parser_creator(gridftp_config_blackpearl)
