import os
from typing import Any, Iterator
from unittest import mock

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem

from gcs_migration.parsers import S3Config


@pytest.fixture
def s3_host_name() -> str:
    return "s3.amazonaws.com"


@pytest.fixture(params=["$HOME/.globus/s3", "/etc/globus/s3-key-map"])
def s3_map_filename(request: Any, connector: str) -> str:
    return request.param


@pytest.fixture(params=[True, False])
def s3_use_https(request: Any, connector: str) -> bool:
    return request.param


@pytest.fixture
def s3_config(s3_host_name: str, s3_map_filename: str, s3_use_https) -> str:
    return f"""\
# Fully-qualified domain name of the S3-compatible service, with a port
# number optionally appended (e.g s3.example.com or s3.example.com:8443).
# Existing buckets must have been created in this region to be accessible.
host_name = {s3_host_name}

# Path to a mime-types file to associate content types with extensions for
# objects stored in this service (usually /etc/mime.types).  If unset,
# all files will be stored with a type of 'binary/octet-stream'.
mime_types_path = /etc/mime.types

# Set to true to use https, or false to use http (default to https)
use_https = {'true' if s3_use_https else 'false'}

# Use virtual-host style (bucket as host name) for S3 requests
# (default to false)
use_virtual_host = false

# Verify https peer credentials
# This requires that the S3 host's ssl certificate be signed by a CA
# included in the operating system's trust root.  To add trusted CA certs,
# see the man page of update-ca-certificates or update-ca-trust, depending
# on the system.
# (default to true)
verify_peer = true

# Maximum number of retries for connection faults
max_retries = 5

# Base retry delay for exponential back-off when retrying (in ms)
retry_delay = 400

# Map filename to map user names to access_key_id and secret_access_key
# The format of the file is
#   username;access_key_id;secret_access_key
# Lines starting with # are ignored.
# The strings $HOME and $LOGNAME are replaced with the current user's
# home directory and user name repectively
s3_map_filename = {s3_map_filename}
"""


@pytest.fixture
def mock_s3_config(s3_host_name: str, s3_map_filename: str) -> S3Config:
    res = mock.MagicMock(spec=S3Config)
    res.path = "/etc/globus/globus-gridftp-server-s3.conf"
    res.host_name = s3_host_name
    res.s3_map_filename = s3_map_filename
    res.use_https = True

    return res


@pytest.fixture
def s3_config_file(fs: FakeFilesystem, s3_config: str) -> Iterator[str]:
    s3_config_file_path = "/etc/globus/globus-gridftp-server-s3.conf"
    fs.makedirs(os.path.dirname(s3_config_file_path))
    with open(s3_config_file_path, "w") as f:
        f.write(s3_config)
    yield s3_config_file_path


@pytest.fixture
def s3_config_object(fs: FakeFilesystem, s3_config_file: str) -> S3Config:
    return S3Config(s3_config_file)
