import os
from typing import Any, Dict, Generator, Optional, Tuple
from unittest import mock

import pytest

from gcs_migration.parsers import GCSConfig

TEST_UUID = "2bbe94f1-3627-475d-a8d7-9482d212dcf0"

GcsConfigFixture = Tuple[str, Dict[str, str]]


@pytest.fixture(params=["/some/other/path", "/etc/globus-connect-server.conf"])
def gcs_config(request: Any, fs: Any) -> Generator[GcsConfigFixture, None, None]:
    dirname = os.path.dirname(request.param)
    if not os.path.exists(dirname):
        os.makedirs(dirname, 0o0755)

    with open(request.param, "w") as f:
        f.write(
            """\
[Globus]
User = user
Password = password
Instance = Sandbox
"""
        )

    yield request.param, {"user": "user", "password": "password", "instance": "Sandbox"}


@pytest.fixture
def endpoint_uuid(fs: Any) -> str:
    gcsdir = "/var/lib/globus-connect-server"
    os.makedirs(gcsdir, mode=0o755)
    with open(f"{gcsdir}/gcsv4-endpoint-id.txt", "w") as f:
        print(TEST_UUID, file=f)

    return TEST_UUID


def test_gcs_config_parser(gcs_config: GcsConfigFixture, endpoint_uuid: str) -> None:
    file_path: Optional[str]

    file_path, contents = gcs_config
    if file_path == "/etc/globus-connect-server.conf":
        file_path = None

    parser = GCSConfig(path=file_path)

    for k, v in contents.items():
        assert v == getattr(parser, k)
    assert parser.endpoint_id == endpoint_uuid


def test_gcs_config_parser_no_uuid_file(gcs_config: GcsConfigFixture) -> None:
    file_path = gcs_config[0]

    with mock.patch("gcs_migration.parsers.gcs_config.os") as mock_os:
        mock_os.path.exists.return_value = False

        parser = GCSConfig(path=file_path)

        assert parser.endpoint_id is None


def test_gcs_config_parser_unreadable_uuid_file(gcs_config: GcsConfigFixture) -> None:
    file_path = gcs_config[0]

    with mock.patch("gcs_migration.parsers.gcs_config.os") as mock_os, mock.patch(
        "gcs_migration.parsers.gcs_config.open"
    ) as mock_open:
        mock_os.path.exists.return_value = True
        mock_open.side_effect = PermissionError

        parser = GCSConfig(path=file_path)

        assert parser.endpoint_id is None


def test_get(gcs_config: GcsConfigFixture) -> None:
    file_path, contents = gcs_config
    parser = GCSConfig(path=file_path)

    assert parser.get("Globus", "Instance") == parser.instance
    assert parser.instance == contents["instance"]
    assert parser.get("Globus", "Instance") == contents["instance"]
